import { globalLookup, globalSetting, signalValues } from "./global";
import { LineDrawer } from './canvas';
import { domContainer, pluginRenderTimeGrid, pluginRenderValues, mountTree,
         genOnWheel, genKeyHandler, keyBindo } from './wave-view';

import { createCodeMirrorState, mountCodeMirror6 } from 'waveql';

let mainRenderEl = null;
const canvasMap = new Map();

/**
 * 
 * @returns {Element}
 */
function getMainRenderEl() {
    if (!mainRenderEl) {
        const el = document.querySelectorAll('.vcd-render-wrapper')[0];
        if (el) {
            mainRenderEl = el;
        }
    }
    return mainRenderEl;
}

/**
 * 
 * @param {{
 *  time: BigInt,
 *  cmd: number,
 *  value?: BigInt,
 *  mask?: BigInt
 * }} dataItem 
 * @returns 
 */
function isBlocked(dataItem) {
    const mask = dataItem.mask;
    if (mask && mask === 1n) {
        return true;
    }
    if (mask && mask === 0n) {
        return false;
    }
    if (mask === undefined && dataItem.cmd > 15) {
        return true;
    }
    return false;
}

/**
 * @param {{
 * kind: 'var' | 'scope',
 * type: 'wire' | 'reg' | 'module',
 * name: string,
 * size: number,
 * link: string
 * }} signal 
 */
function makeWaveCanvas(signal) {
    const mainRender = getMainRenderEl();
    const canvas = document.createElement('canvas');
    canvas.id = 'wave-' + signal.link;
    canvas.className = 'render-canvas';
    canvas.width = document.body.clientWidth;
    canvas.height = globalSetting.displaySignalHeight;
    
    mainRender.appendChild(canvas);
    canvasMap.set(signal, canvas);

    const ctx = canvas.getContext('2d');
    const datas = signalValues[signal.link];

    if (signal.size === 1) {
        drawSimpleSignal(ctx, datas);
    } else {
        // adwa
    }
}

/**
 * @param {{
* kind: 'var' | 'scope',
* type: 'wire' | 'reg' | 'module',
* name: string,
* size: number,
* link: string
* }} signal 
*/
function makeWaveSvg(signal) {
    const mainRender = getMainRenderEl();
    const height = globalSetting.displaySignalHeight;
    const width = document.body.clientWidth;
    const draw = SVG().addTo(mainRender).size(width, height);
    const datas = signalValues[signal.link];

    if (signal.size === 1) {
        drawSimpleSignalSvg(draw, datas);
    } else {
        // adwa
    }
}


/**
 * 
 * @param {Svg} ctx 
 * @param {Array<{
*  time: BigInt,
*  cmd: number,
*  value?: BigInt,
*  mask?: BigInt
* }>} signals 
*/
function drawSimpleSignalSvg(ctx, signals) {
   let lastData = null;
   let lastX = 300;
   let lastColor = '';

//    ctx.lineWidth = 3;
//    ctx.lineJoin = 'round';
   
//    const drawer = new LineDrawer(ctx);
//    drawer.beginPath();
   for (const data of signals) {
       if (lastData === null) {
            lastData = data;
            lastX = 300;
            lastColor = getDataRenderColor(data);
            let lastY = Math.max((15 - data.cmd) * 30, 0) + 10;
        // drawer.moveTo(last_x, last_y);
            ctx.move(lastX, lastY);

       } else {
           // 高阻态
           const start = lastData.time;
           const end = data.time;
           console.log(lastData, isBlocked(lastData));
           let lastY = Math.max((15 - lastData.cmd) * 30, 0) + 10;

           const duration = parseInt(end - start);
           const currentX = lastX + 100;
           const currentColor = getDataRenderColor(data);
           console.log(currentX, lastY, duration);

           const line = ctx.line(lastX, lastY, currentX, lastY);

           if (currentColor !== lastColor) {
                ctx.stroke({ color: lastColor, linecap: 'round', width: 3 });
           }
        //    drawer.lineTo(current_x, last_y, color);

           if (lastData.cmd !== data.cmd) {
               let currentY = Math.max((15 - data.cmd) * 30, 0) + 10;
               const color = getDataRenderColor(data);
               ctx.line(currentX, lastY, currentX, currentY);
            //    drawer.lineTo(current_x, current_y, color);
           }

           lastData = data;
           lastX = currentX;
           lastColor = currentColor;
       }
   }

   // 如果 last_data 的时间不足最长时间，就用最后一个时刻的去补足
   ctx.stroke({ color: lastColor, linecap: 'round', width: 3 });
}


function getDataRenderColor(data) {
    return isBlocked(data) ? 'rgb(224,108,117)' : 'rgb(10, 200, 10)';
}


/**
 * @param {{
* kind: 'var' | 'scope',
* type: 'wire' | 'reg' | 'module',
* name: string,
* size: number,
* link: string
* }} signal
*/
function removeWaveCanvas(signal) {
    const canvas = canvasMap.get(signal);
    if (canvas instanceof HTMLCanvasElement) {
        canvas.parentNode.removeChild(canvas);
    }
}


/**
 * 
 * @param {CanvasRenderingContext2D} ctx 
 * @param {Array<{
 *  time: BigInt,
 *  cmd: number,
 *  value?: BigInt,
 *  mask?: BigInt
 * }>} signals 
 */
function drawSimpleSignal(ctx, signals) {
    let last_data = null;
    let last_x = 300;

    ctx.lineWidth = 3;
    ctx.lineJoin = 'round';
    const drawer = new LineDrawer(ctx);
    
    drawer.beginPath();
    for (const data of signals) {
        if (last_data === null) {
            last_data = data;
            last_x = 300;
            let last_y = Math.max((15 - last_data.cmd) * 30, 0) + 10;
            drawer.moveTo(last_x, last_y);
        } else {
            // 高阻态
            const start = last_data.time;
            const end = data.time;
            console.log(last_data, isBlocked(last_data));
            let last_y = Math.max((15 - last_data.cmd) * 30, 0) + 10;

            const duration = parseInt(end - start);
            const current_x = last_x + 100;

            console.log(current_x, last_y, duration);

            const color = getDataRenderColor(last_data);      
            drawer.lineTo(current_x, last_y, color);

            if (last_data.cmd !== data.cmd) {
                let current_y = Math.max((15 - data.cmd) * 30, 0) + 10;
                const color = getDataRenderColor(data);
                drawer.lineTo(current_x, current_y, color);
            }

            last_data = data;
            last_x = current_x;
        }
    }

    // 如果 last_data 的时间不足最长时间，就用最后一个时刻的去补足
    
    ctx.stroke();
}

function pluginLocalStore(desc, pstate /* , els */) {
    const pstateJsonString = JSON.stringify({
        // yOffset: pstate.yOffset,
        xOffset: pstate.xOffset,
        xScale: pstate.xScale
    });
    localStorage.setItem('dide', pstateJsonString);
}

/**
 * 
 * @param {HTMLElement} parentElement 
 */
function makeWaveView(parentElement) {
    if (!parentElement) {
        parentElement = getMainRenderEl();
    }
    const container = domContainer({
        elemento: mountTree.defaultElemento,
        layers: mountTree.defaultLayers,
        renderPlugins: [
            pluginRenderTimeGrid,
            pluginRenderValues,
            pluginLocalStore
        ]
    });

    parentElement.appendChild(container.pstate.container);
    container.start(globalLookup);

    globalLookup.hasHistory = true;
    globalLookup.isRO = true;
    globalLookup.updater = ( /* str */ ) => {
        console.log('updater');
    };

    const cmState = createCodeMirrorState(globalLookup, container.pstate);

    const cm = mountCodeMirror6(
        cmState,
        container.elo.waveqlPanel,
        globalLookup,
        container.pstate
    );

    cm.view.dispatch({changes: {from: 0, insert: ' '}});
    cm.view.dispatch({changes: {from: 0, to: 1, insert: ''}});

    container.elo.container.addEventListener('keydown', genKeyHandler.genKeyHandler(parentElement, container.pstate, globalLookup, cm, keyBindo));
    container.elo.container.addEventListener('wheel', genOnWheel(parentElement, container.pstate, globalLookup, cm, keyBindo));
    // console.log(cm);
    cm.view.focus();
}


/**
 * @param {{
* kind: 'var' | 'scope',
* type: 'wire' | 'reg' | 'module',
* name: string,
* size: number,
* link: string
* }} signal
*/
function toggleRender(signal) {
    if (globalLookup.currentWires.has(signal)) {
        globalLookup.currentWires.delete(signal);
        // removeWaveCanvas(signal);
        globalLookup.render();
    } else {
        // makeWaveSvg(signal);
        const lane = { ref: signal.link };
        
        globalLookup.currentWires.add(signal);
        globalLookup.view.push(lane);
        globalLookup.render();
    }
}

export {
    getMainRenderEl,
    makeWaveCanvas,
    makeWaveSvg,
    removeWaveCanvas,
    makeWaveView,
    toggleRender
}